import logging

from nicegui import ui

logger = logging.getLogger("xmas_app")


async def initialize_qwebchannel() -> None:
    """Initializes a reusable `QWebChannel` including a handler."""
    success, error = await ui.run_javascript(
        """
            return (function () {
                if (window.ensureXmasHandler) {
                    return [true, null];
                }

                if (!window.qt || !window.QWebChannel || !qt.webChannelTransport) {
                    return [false, 'Qt objects missing'];
                }

                window.ensureXmasHandler = async function ensureXmasHandler() {
                    if (window.xmasHandler) {
                        return window.xmasHandler;
                    }

                    return await new Promise((resolve, reject) => {
                        new QWebChannel(qt.webChannelTransport, function (channel) {
                            window.xmasChannel = channel;
                            window.xmasHandler = channel.objects.handler;
                            resolve(window.xmasHandler);
                        });
                    });
                };
                return [true, null];
            })();
        """
    )
    if not success:
        logger.error("failed to initialize QWebChannel: %s", error)
        ui.notify("Kommunikation mit QGIS nicht möglich", type="negative")


async def connect_signal(signal_name: str) -> None:
    """Connects NiceGUI's `emitEvent` function to a `QWebChannel` handler signal."""
    success, error = await ui.run_javascript(
        f"""
            return (async function () {{
                if (!window.ensureXmasHandler) {{
                    return [false, 'ensureXmasHandler missing'];
                }}

                const handler = window.xmasHandler || await window.ensureXmasHandler().catch(err => {{
                    return [false, err];
                }});

                const callback = handler.{signal_name}Callback || (handler.{signal_name}Callback = (data) => emitEvent('{signal_name}', data));
                try {{
                    handler.{signal_name}.disconnect(callback);
                }} catch (err) {{
                    // handler may not have been connected yet
                }}
                handler.{signal_name}.connect(callback);
                return [true, null];
            }})();
        """
    )
    if not success:
        ui.notify(f"Verbindung zu Signal {signal_name} fehlgeschlagen", type="negative")
        logger.error(
            "failed to bind %r signal to QWebChannel handler: %s",
            signal_name,
            error,
        )



import os
from functools import cached_property, lru_cache
from pathlib import Path
from typing import Literal, Optional, Self

from pydantic import HttpUrl, model_validator
from pydantic_extra_types.semantic_version import SemanticVersion
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)
from xplan_tools.interface.db import DBRepository
from xplan_tools.model.base import Appschema

from xmas_app.models.mappings import AssociationTableMapping


class Settings(BaseSettings):
    debug: bool = False

    PGUSER: Optional[str] = None
    PGPASSWORD: Optional[str] = None
    PGHOST: Optional[str] = None
    PGPORT: Optional[str] = None
    PGDATABASE: Optional[str] = None
    PGSERVICE: Optional[str] = None

    appschema: str
    appschema_version: str
    appschema_cls: Appschema | None = None
    app_port: int
    db_type: str = "postgres"  # defaulting to postgres
    app_mode: Literal["dev", "prod"] = "prod"
    codelist_repo: HttpUrl = HttpUrl(
        "https://registry.gdi-de.org/codelist/de.xleitstelle.xplanung"
    )
    qgis_plugin_name: str = "XMAS-Plugin"
    qgis_plugin_min_version: SemanticVersion = SemanticVersion(major=0, minor=17)
    mappings_toml: str = str(Path(__file__).parent / "resources/mappings.toml")

    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8")

    @model_validator(mode="after")
    def _after_validator(self) -> Self:
        self.appschema_cls = Appschema.from_prefix(
            self.appschema, self.appschema_version
        )
        self.repo
        return self

    @cached_property
    def repo(self):
        """Set environment variable, initialize and return repo."""
        os.environ.update(
            self.model_dump(
                include={
                    "PGUSER",
                    "PGHOST",
                    "PGPASSWORD",
                    "PGPORT",
                    "PGDATABASE",
                    "PGSERVICE",
                },
                exclude_none=True,
            )
            | {"PGCONNECT_TIMEOUT": "5"}  # add a default timeout
        )
        return DBRepository(
            "postgresql://",
        )


@lru_cache
def get_settings() -> Settings:
    return Settings()


class Mappings(BaseSettings):
    association_table: AssociationTableMapping

    model_config = SettingsConfigDict(toml_file=get_settings().mappings_toml)

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (TomlConfigSettingsSource(settings_cls),)


@lru_cache
def get_mappings() -> Mappings:
    return Mappings()



import logging

from fastapi import HTTPException
from nicegui import run
from pydantic import ValidationError
from xplan_tools.model import model_factory

from xmas_app.models.crud import InsertPayload, UpdatePayload
from xmas_app.settings import get_settings

logger = logging.getLogger("xmas_app")


async def create(payload: InsertPayload, plan_id: str):
    """Create features in the database.

    Iterates over data items in the payload and validates them with pydantic.

    For non-geometry aka text features, the corresping plan is updated to have
    a reference to the new text feature so that there are no orphans in the db.

    If everything is valid, the features and feature update are saved to the database.
    """
    logger.debug(
        "Processing create payload for plan with ID '%s': %s", plan_id, payload
    )
    features = []
    feature_updates = []
    errors = []
    for item in payload.root:
        try:
            featuretype = model_factory(item.featuretype, item.version, item.appschema)
            geom_field = featuretype.get_geom_field()
            properties = item.properties
            properties.pop("featuretype", None)
            if item.geometry:
                properties[geom_field] = item.geometry
            feature = featuretype.model_validate(properties)
            features.append(feature)
            # if feature has geom_field, i.e. it's not a text feature, continue
            if geom_field:
                continue
            # else add unidirectional reference from corresponding plan to text feature
            try:
                plan = get_settings().repo.get(plan_id)
                plan_data = plan.model_dump(mode="json")
            except Exception as e:
                logger.exception("failed to load plan: %s", e)
            else:
                for assoc in plan.get_associations():
                    prop_info = plan.get_property_info(assoc)
                    if item.featuretype not in prop_info["typename"]:
                        continue
                    ref = plan_data[assoc] or ([] if prop_info["list"] else "")
                    if feature.id in ref:
                        continue
                    if prop_info["list"]:
                        ref.append(feature.id)
                    else:
                        ref = feature.id
                    plan_data[assoc] = ref
                    break
                feature_updates.append(plan.model_validate(plan_data))

        except ValidationError as e:
            errors.append(e.errors())
    if errors:
        raise HTTPException(
            status_code=422,
            detail=errors,
        )
    try:
        await run.io_bound(get_settings().repo.save_all, features)
        if feature_updates:
            await run.io_bound(get_settings().repo.update_all, feature_updates)
    except Exception as e:
        logger.exception("Exception occured while saving features to db: %s", e)
        raise HTTPException(
            status_code=500,
            detail=str(e),
        )


async def update(payload: UpdatePayload):
    """Update features in the database.

    Iterates over data items in the payload and updates existing features with them.

    The update data is validated with pydantic.

    If everything is valid, the features are saved to the database.
    """
    logger.debug("Processing update payload: %s", payload)
    features = []
    errors = []
    for id, item in payload.root.items():
        feature = await run.io_bound(get_settings().repo.get, str(id))
        update = item.properties
        if item.geometry:
            update[feature.get_geom_field()] = item.geometry
        data = feature.model_dump() | update
        try:
            feature = feature.model_validate(data)
            features.append(feature)
        except ValidationError as e:
            errors.append(e.errors())
    if errors:
        raise HTTPException(
            status_code=422,
            detail=errors,
        )
    try:
        await run.io_bound(get_settings().repo.update_all, features)
    except Exception as e:
        logger.exception("Exception occured while updating features in db: %s", e)
        raise HTTPException(
            status_code=500,
            detail=str(e),
        )



import json
import logging
from typing import Literal, get_args

from nicegui import events, ui

logger = logging.getLogger("xmas_app")


class FeatureInteractionButton(ui.button):
    """Generic Button to interact with QGIS features via QWebchannel."""

    action_types = Literal["highlight_feature", "show_attribute_form", "select_feature"]
    type_map = {
        "highlight_feature": {"text": "Zum Feature springen", "icon": "o_pageview"},
        "show_attribute_form": {
            "text": "Attributformular öffnen",
            "icon": "list_alt",
        },
        "select_feature": {"text": "Feature selektieren", "icon": "select_all"},
    }

    def __init__(
        self,
        action_type: action_types,
        target: str,
        source: str | None = None,
        geom: bool = True,
    ):
        """Button initialization.

        Args:
            type: the specific action to initialize the button for
            target: the ID of the target feature
            source: the ID of the source feature
            geom: whether the feature has a geometry
        Raises:
            ValueError: if the type is unknown
        """
        if action_type not in get_args(self.action_types):
            raise ValueError(f"invalid type: '{action_type}'")
        self.action_type = action_type
        self.source = source
        self.target = target
        super().__init__(**self.type_map[action_type], on_click=self._on_click)
        self.props("flat square no-caps align=left")
        self.classes("w-full")
        if not geom:
            self.disable()
            self.tooltip("Keine Geometrie")

    async def _on_click(self, _: events.ClickEventArguments):
        try:
            logger.debug("Sending %s request to QWebChannel handler", self.action_type)
            data = json.dumps({"source": self.source, "target": self.target})
            ui.run_javascript(f"""new QWebChannel(qt.webChannelTransport, function (channel) {{
                                        channel.objects.handler.{self.action_type}({data})
                                }});""")
        except Exception as e:
            logger.exception(
                "Exception while %s was called: %s", self.action_type, str(e)
            )
